from nano_settings.code import BaseConfig
from nano_settings.code import ConfigValidationError
from nano_settings.code import EnvAlias
from nano_settings.code import SecretStr
from nano_settings.code import from_env
from nano_settings.code import looks_like_boolean



"""Tests."""

from dataclasses import dataclass
import json
from typing import Annotated
from typing import Any
from unittest import mock
from unittest.mock import patch

import pytest

from nano_settings.code import BaseConfig
from nano_settings.code import ConfigValidationError
from nano_settings.code import EnvAlias
from nano_settings.code import SecretStr
from nano_settings.code import from_env
from nano_settings.code import looks_like_boolean


def test_base_config_easy():
    """Must create config instance from env vars."""

    # arrange
    @dataclass
    class EasyConfig(BaseConfig):
        variable_1: int
        variable_2: str
        variable_3: bool = False

    # act
    with patch.dict(
        'os.environ',
        EASYCONFIG__VARIABLE_1='1',
        EASYCONFIG__VARIABLE_2='string',
    ):
        config = from_env(EasyConfig)

    # assert
    assert config.variable_1 == 1
    assert config.variable_2 == 'string'
    assert not config.variable_3


def test_base_config_skipped_unset():
    """Must fail to create because we have unset parameter."""
    # arrange
    output = mock.Mock()

    @dataclass
    class BadConfig(BaseConfig):
        _variable_0: int
        variable_1: int
        variable_2: str
        variable_3: bool = False

    # act
    with (
        patch.dict(
            'os.environ',
            BADCONFIG__VARIABLE_1='1',
            BADCONFIG__VARIABLE_2='string',
        ),
        pytest.raises(SystemExit),
    ):
        from_env(BadConfig, output=output)

    # assert
    output.assert_has_calls(
        [
            mock.call(
                "Field '_variable_0' is supposed to have a default value"
            ),
        ]
    )


def test_base_config_skipped():
    """Must create config with skipped env variables."""
    # arrange
    output = mock.Mock()

    @dataclass
    class EasyConfig(BaseConfig):
        variable_1: int
        variable_2: str
        _variable_3: int = 0
        _variable_4: str = 'a'

    # act
    with patch.dict(
        'os.environ',
        EASYCONFIG__VARIABLE_1='1',
        EASYCONFIG__VARIABLE_2='string',
    ):
        config = from_env(EasyConfig, output=output)

    # assert
    assert config.variable_1 == 1
    assert config.variable_2 == 'string'
    assert config._variable_3 == 0
    assert config._variable_4 == 'a'


def test_base_config_medium():
    """Must create config instance from env vars."""

    # arrange
    @dataclass
    class MediumConfig(BaseConfig):
        variable_1: Annotated[int, int]
        variable_2: Annotated[str, str.title]
        variable_3: Annotated[bool, looks_like_boolean]

    # act
    with patch.dict(
        'os.environ',
        MEDIUMCONFIG__VARIABLE_1='1',
        MEDIUMCONFIG__VARIABLE_2='string',
        MEDIUMCONFIG__VARIABLE_3='true',
    ):
        config = from_env(MediumConfig)

    # assert
    assert config.variable_1 == 1
    assert config.variable_2 == 'String'
    assert config.variable_3


def test_base_config_hard():
    """Must create config instance from env vars."""
    # arrange

    @dataclass
    class Database(BaseConfig):
        url: str
        timeout: int

    @dataclass
    class Email(BaseConfig):
        email: str
        retries: bool = False

    @dataclass
    class HardConfig(BaseConfig):
        variable_1: int
        email: Email
        database: Database

    # act
    with patch.dict(
        'os.environ',
        HARDCONFIG__VARIABLE_1='1',
        HARDCONFIG__EMAIL__EMAIL='john@snow.com',
        HARDCONFIG__DATABASE__URL='https://site.com',
        HARDCONFIG__DATABASE__TIMEOUT='1',
    ):
        config = from_env(HardConfig)

    # assert
    assert config.variable_1 == 1
    assert config.email.email == 'john@snow.com'
    assert config.database.url == 'https://site.com'
    assert config.database.timeout == 1


def test_base_config_union():
    """Must fail to create config because of the Union type."""
    # arrange
    output = mock.Mock()

    @dataclass
    class BadConfig(BaseConfig):
        variable_1: int | None
        variable_2: str

    # act + assert
    with pytest.raises(SystemExit):
        from_env(BadConfig, output=output)

    output.assert_has_calls(
        [
            mock.call(
                'Config values are not supposed to be '
                'of Union type: variable_1: int | None'
            ),
            mock.call(
                "Environment variable 'BADCONFIG__VARIABLE_2' is not set"
            ),
        ]
    )


def test_secret_str_len():
    """Must use magic method of the object."""
    # arrange
    reference = 'hello world'

    # act
    target = SecretStr(reference)

    # assert
    assert len(reference) == len(target)


def test_secret_str_test_1():
    """Must use magic method of the object."""
    # arrange
    reference = 'hello world'

    # act
    target = SecretStr(reference)

    # assert
    assert str(target) == repr(target) != reference


def test_secret_str_test_2():
    """Must use magic method of the object."""
    # arrange
    reference = 'hello world'

    # act
    target = SecretStr(reference)

    # assert
    assert set(str(target)) == {'*'}


def test_secret_str_get():
    """Must use magic method of the object."""
    # arrange
    reference = 'hello world'

    # act
    target = SecretStr(reference)

    # assert
    assert target != reference
    assert target.get_secret_value() == reference


def test_base_config_wrong_type():
    """Must fail to create config because wrong type is used."""
    # arrange
    output = mock.Mock()

    @dataclass
    class BadConfig(BaseConfig):
        variable_1: Annotated[str, int]

    # act
    with (
        patch.dict(
            'os.environ',
            BADCONFIG__VARIABLE_1='test',
        ),
        pytest.raises(SystemExit),
    ):
        from_env(BadConfig, output=output)

    # assert
    output.assert_has_calls(
        [
            mock.call(
                "Failed to convert 'variable_1' to type 'str', "
                'got ValueError: invalid literal for'
                " int() with base 10: 'test'"
            ),
        ]
    )


def test_base_config_wrong_logic():
    """Must fail to create config because wrong value is used."""
    # arrange
    output = mock.Mock()

    def bigger_than_one(value: Any) -> Any:
        """Raise if we've got wrong value."""
        if int(value) <= 1:
            msg = 'Not bigger'
            raise ConfigValidationError(msg)

    @dataclass
    class BadConfig(BaseConfig):
        variable_1: Annotated[str, bigger_than_one]

    # act
    with (
        patch.dict(
            'os.environ',
            BADCONFIG__VARIABLE_1='1',
        ),
        pytest.raises(SystemExit),
    ):
        from_env(BadConfig, output=output)

    # assert
    output.assert_has_calls([mock.call('Not bigger')])


def test_base_config_alias_good():
    """Must find variable using different name."""
    # arrange

    @dataclass
    class GoodConfig(BaseConfig):
        variable_1: Annotated[str, EnvAlias('MY_OTHER_VARIABLE')]

    # act
    with patch.dict(
        'os.environ',
        MY_OTHER_VARIABLE='1',
    ):
        config = from_env(GoodConfig)

    # assert
    assert config.variable_1 == '1'


def test_base_config_alias_bad():
    """Must fail to create config because no env variables are set."""
    # arrange
    output = mock.Mock()

    @dataclass
    class BadConfig(BaseConfig):
        variable_1: Annotated[str, EnvAlias('_A', '_B', '_C')]

    # act
    with pytest.raises(SystemExit):
        from_env(BadConfig, output=output)

    # assert
    output.assert_has_calls(
        [
            mock.call(
                'None of expected environment variables are set: '
                "'BADCONFIG__VARIABLE_1', '_A', '_B', '_C'"
            )
        ]
    )


def test_base_config_default():
    """Must create config using default."""
    # arrange
    output = mock.Mock()

    @dataclass
    class GoodConfig(BaseConfig):
        variable_1: Annotated[tuple, json.loads] = ('key',)

    # act
    config = from_env(GoodConfig, output=output)

    # assert
    assert config.variable_1 == ('key',)


def test_base_config_long_conversion():
    """Must create config using more than one step of casting."""
    # arrange
    output = mock.Mock()
    reference = 2

    @dataclass
    class GoodConfig(BaseConfig):
        variable_1: Annotated[
            tuple,
            int,
            float,
            lambda x: x['value'],
            lambda x: x['key'],
            json.loads,
        ]

    # act
    with patch.dict(
        'os.environ',
        GOODCONFIG__VARIABLE_1='{"key": {"value": 2.55}}',
    ):
        config = from_env(GoodConfig, output=output)

    # assert
    assert config.variable_1 == reference



"""Simple config class, when you do not need full size pydantic."""

from collections.abc import Callable
from dataclasses import MISSING
from dataclasses import Field
from dataclasses import dataclass
from dataclasses import fields
import os
import sys
import types
from typing import Annotated
from typing import Any
from typing import TypeVar
from typing import get_args
from typing import get_origin

__all__ = [
    'BaseConfig',
    'ConfigValidationError',
    'EnvAlias',
    'SecretStr',
    'from_env',
    'looks_like_boolean',
]


@dataclass
class BaseConfig:
    """Configuration base class."""


class ConfigValidationError(Exception):
    """Failed to cast attribute to expected type."""


class SecretStr:
    """String that does not show its value."""

    def __init__(self, secret_value: str) -> None:
        """Initialize instance."""
        self._secret_value = secret_value

    def get_secret_value(self) -> str:
        """Return secret value."""
        return self._secret_value

    def __len__(self) -> int:
        """Return number of symbols."""
        return len(self._secret_value)

    def __str__(self) -> str:
        """Return string representation."""
        return '**********' if self.get_secret_value() else ''

    def __repr__(self) -> str:
        """Return string representation."""
        return self.__str__()


class EnvAlias:
    """Alternative name or names of environment variable for a field.

    Alias is expected to be the rightmost element in Annotated hint.
    """

    def __init__(self, *names: str) -> None:
        """Initialize instance."""
        self.names = names

    def __call__(self, value: str | None) -> str | None:
        """Do nothing, juts return."""
        return value

    def find_matching(self, env_name: str) -> tuple[str | None, str | None]:
        """Try getting value via another name."""
        for name in self.names:
            value = os.environ.get(name)

            if value is not None:
                return value, None

        variables = ', '.join(map(repr, (env_name, *self.names)))
        return (
            None,
            f'None of expected environment variables are set: {variables}',
        )


T_co = TypeVar('T_co', bound=BaseConfig, covariant=True)


def looks_like_boolean(value: str) -> bool:
    """Return True if value looks like boolean."""
    return value.lower() == 'true'


def _is_excluded(field: Field, field_exclude_prefix: str) -> bool:
    """Return true if this field is excluded from env vars."""
    return bool(
        field_exclude_prefix and field.name.startswith(field_exclude_prefix)
    )


def _has_no_default(field: Field) -> str | None:
    """Return error message if field expects a value."""
    if field.default is MISSING:
        return f'Field {field.name!r} is supposed to have a default value'
    return None


def _is_union(field: Field) -> str | None:
    """Return error message if field is a Union type."""
    if isinstance(field.type, types.UnionType):
        return (
            f'Config values are not supposed '
            f'to be of Union type: {field.name}: {field.type}'
        )
    return None


def from_env(  # noqa: C901, PLR0912, PLR0915
    model_type: type[T_co],
    *,
    env_prefix: str = '',
    env_separator: str = '__',
    field_exclude_prefix: str = '_',
    output: Callable = print,
    _prefixes: tuple[str, ...] | None = None,
    _terminate: Callable = lambda: sys.exit(1),
) -> T_co:
    """Build instance from environment variables."""
    errors: list[str] = []
    attributes: dict[str, Any] = {}

    if _prefixes is None:
        env_prefix = env_prefix or model_type.__name__.upper()
        _prefixes = _prefixes or (env_prefix, env_separator)

    for field in fields(model_type):
        if _is_excluded(field, field_exclude_prefix):
            msg = _has_no_default(field)
            if msg:
                errors.append(msg)
            continue

        msg = _is_union(field)
        if msg:
            errors.append(msg)
            continue

        if get_origin(field.type) is Annotated:
            expected_type, *casting_callables = get_args(field.type)
        else:
            expected_type = field.type
            casting_callables = [field.type]

        if issubclass(expected_type, BaseConfig):
            value = from_env(
                model_type=expected_type,
                env_prefix='',
                field_exclude_prefix=field_exclude_prefix,
                output=output,
                _prefixes=(*_prefixes, field.name.upper(), env_separator),
                _terminate=_terminate,
            )
            casting_callables.pop()

        else:
            prefix = ''.join(_prefixes)
            env_name = f'{prefix}{field.name}'.upper()
            value = os.environ.get(env_name)

            if value is None and field.default is not MISSING:
                # using default without data casting
                value = field.default
                casting_callables = []

            if value is None and isinstance(casting_callables[-1], EnvAlias):
                value, msg = casting_callables[-1].find_matching(env_name)

                if msg:
                    errors.append(msg)
                    continue

            if value is None:
                msg = f'Environment variable {env_name!r} is not set'
                errors.append(msg)
                continue

        final_value = value
        for _callable in reversed(casting_callables):
            try:
                final_value = _callable(final_value)
            except ConfigValidationError as exc:
                errors.append(str(exc))
                break
            except Exception as exc:
                msg = (
                    f'Failed to convert {field.name!r} '
                    f'to type {expected_type.__name__!r}, '
                    f'got {type(exc).__name__}: {exc}'
                )
                errors.append(msg)
                break

        attributes[field.name] = final_value

    if errors:
        for error in errors:
            output(error)
        _terminate()

    return model_type(**attributes)
